/* vi: ts=8 sts=4 sw=4
 *
 * This file is part of the KDE project, module kdesu.
 * Copyright (C) 1999,2000 Geert Jansen <g.t.jansen@stud.tue.nl>
 *
 * secure.cpp: Peer credentials for a UNIX socket.
 */

#include <config.h>

#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <errno.h>

#if defined (SOLARIS10)
#include <ucred.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>

#include <kdebug.h>
#include "secure.h"


/**
 * Under Linux, Socket_security is supported.
 */

#if defined(SO_PEERCRED)

SocketSecurity::SocketSecurity(int sockfd)
    : ok(false)
{
    ksize_t len = sizeof(struct ucred);
    if (getsockopt(sockfd, SOL_SOCKET, SO_PEERCRED, &cred, &len) < 0) {
	kdError() << "getsockopt(SO_PEERCRED) " << perror << endl;
	return;
    }
    ok = true;
}

#else
# if defined(HAVE_GETPEEREID) || defined (SOLARIS10)
SocketSecurity::SocketSecurity(int sockfd)
{
#if defined (SOLARIS10)
    ucred_t* ucr = NULL;

    ucr = ucred_get(getpid());
    if (ucr != NULL)
    {
	cred.uid = ucred_geteuid(ucr);
	cred.gid = ucred_getegid(ucr);
	cred.pid = -1;
	ok = true;
	ucred_free(ucr);
    }
#else
    uid_t euid;
    gid_t egid;
    if (getpeereid(sockfd, &euid, &egid) == 0) {
	cred.uid = euid;
	cred.gid = egid;
	cred.pid = -1;
	ok = true;
    }
#endif
}

# else

/**
 * The default version does nothing.
 */

SocketSecurity::SocketSecurity(int sockfd)
{
    static bool warned_him = FALSE;

    if (!warned_him) {
        kdWarning() << "Using void socket security. Please add support for your" << endl;
        kdWarning() << "platform to kdesu/kdesud/secure.cpp" << endl;
        warned_him = TRUE;
    }

    // This passes the test made in handler.cpp
    cred.uid = geteuid();
    cred.gid = getegid();
    cred.pid = getpid();
    ok = true;
}

# endif
#endif
