/*****************************************************************

Copyright (c) 1996-2002 the kicker authors. See file AUTHORS.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************/

#ifdef FD_SETSIZE
#undef FD_SETSIZE
#endif
#define FD_SETSIZE 8192

#include <config.h>
#include <klocale.h>
#include <kcmdlineargs.h>
#include <kdebug.h>
#include <kaboutdata.h>
#include <kconfig.h>
#include <dcopclient.h>

#include <X11/Xlib.h>

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <strings.h>
#include <sys/types.h>
#include <sys/resource.h>
#include <errno.h>

#include "kicker.h"

extern "C" {
    static void setSystemLimits (void)
    {
        static struct rlimit maxLimit;
        (void) memset (&maxLimit, '\0', sizeof (struct rlimit));

	if (getrlimit (RLIMIT_NOFILE, &maxLimit) < 0)
	{
	    perror ("getrlimit"), strerror (errno);
	    return;
	}

	if (maxLimit.rlim_cur < 1024)
	    maxLimit.rlim_cur = 1024;
	if (maxLimit.rlim_max < 8192)
	    maxLimit.rlim_max = 8192;

        if (setrlimit (RLIMIT_NOFILE, &maxLimit) < 0)
            perror ("setrlimit"), strerror (errno);
        return;
    }         
}
  
extern "C"
{
    typedef void SignalHandlerFunction (int);
    static SignalHandlerFunction* KSignal (int signo, SignalHandlerFunction* sigFunc)
    {
        static struct sigaction act;
	static struct sigaction oact;
        act.sa_handler = sigFunc;
        sigemptyset (&act.sa_mask);
        act.sa_flags = 0;
        if (signo == SIGALRM)
        {
#ifdef SA_INTERRUPT
            act.sa_flags |= SA_INTERRUPT;
#endif
        }
        else
        {
            act.sa_flags |= SA_RESTART;
        }
        if (sigaction (signo, &act, &oact) != 0)
                return (SIG_ERR);
        return (oact.sa_handler);
    }
}

extern "C"
{
    static void KSignalHandler (int signo)
    {
	KSignal (signo, SIG_IGN);
	KSignal (signo, KSignalHandler);
    }
}


int kicker_screen_number = 0;

static const char description[] =
        I18N_NOOP("The KDE panel");

static const char version[] = VERSION;

static void sighandler(int)
{
    fprintf(stderr, "kicker: sighandler called\n");
    QApplication::exit();
}

extern "C" KDE_EXPORT int kdemain( int argc, char ** argv )
{
    setSystemLimits();
    KSignal(SIGPIPE, KSignalHandler);
    KSignal(SIGALRM, KSignalHandler);
    KSignal (SIGHUP, KSignalHandler);
    KSignal (SIGINT, KSignalHandler);
    KSignal (SIGTERM, KSignalHandler);

    {
        QCString multiHead = getenv("KDE_MULTIHEAD");
        if (multiHead.lower() == "true") {
	    Display *dpy = XOpenDisplay(NULL);
	    if (! dpy) {
		fprintf(stderr, "%s: FATAL ERROR: couldn't open display %s\n",
			argv[0], XDisplayName(NULL));
		exit(1);
	    }

	    int number_of_screens = ScreenCount(dpy);
	    kicker_screen_number = DefaultScreen(dpy);
	    int pos;
	    QCString display_name = XDisplayString(dpy);
	    XCloseDisplay(dpy);
	    dpy = 0;

	    if ((pos = display_name.findRev('.')) != -1)
		display_name.remove(pos, 10);

            QCString env;
	    if (number_of_screens != 1) {
		for (int i = 0; i < number_of_screens; i++) {
		    if (i != kicker_screen_number && fork() == 0) {
			kicker_screen_number = i;
			// break here because we are the child process, we don't
			// want to fork() anymore
			break;
		    }
		}

		env.sprintf("DISPLAY=%s.%d", display_name.data(), kicker_screen_number);

		if (putenv(strdup(env.data()))) {
		    fprintf(stderr,
			    "%s: WARNING: unable to set DISPLAY environment variable\n",
			    argv[0]);
		    perror("putenv()");
		}
	    }
	}
    }

    QCString appname;
    if (kicker_screen_number == 0)
	appname = "kicker";
    else
	appname.sprintf("kicker-screen-%d", kicker_screen_number);

    KAboutData aboutData( appname.data(), I18N_NOOP("KDE Panel"),
                          version, description, KAboutData::License_BSD,
                          I18N_NOOP("(c) 1999-2004, The KDE Team") );

    aboutData.addAuthor("Aaron J. Seigo", I18N_NOOP("Current maintainer"), "aseigo@kde.org");
    aboutData.addAuthor("Matthias Elter",0, "elter@kde.org");
    aboutData.addAuthor("Matthias Ettrich",0, "ettrich@kde.org");
    aboutData.addAuthor("Wilco Greven",0, "greven@kde.org");
    aboutData.addAuthor("Rik Hemsley",0, "rik@kde.org");
    aboutData.addAuthor("Daniel M. Duley",0, "mosfet@kde.org");
    aboutData.addAuthor("Preston Brown",0, "pbrown@kde.org");
    aboutData.addAuthor("John Firebaugh",0, "jfirebaugh@kde.org");
    aboutData.addAuthor("Waldo Bastian", I18N_NOOP("Kiosk mode"), "bastian@kde.org");

    aboutData.addCredit("Jessica Hall", /* I18N_NOOP("KConfigXT") */ 0, "jes.hall@kdemail.net");
    aboutData.addCredit("Stefan Nikolaus", /* I18N_NOOP("Bug fixes") */ 0, "stefan.nikolaus@kdemail.net");
    KCmdLineArgs::init( argc, argv, &aboutData );

    if (!Kicker::start()) {
	kdError() << "kicker is already running!" << endl;
	return 0;
    }

    Kicker* kicker = new Kicker();
    int rv = kicker->exec();
    delete kicker;
    return rv;
}

