/*
    KSysGuard, the KDE System Guard
   
	Copyright (c) 1999, 2000 Chris Schlaeger <cs@kde.org>

	Solaris support by Torsten Kasch <tk@Genetik.Uni-Bielefeld.DE>
  Rewritten by Stefan Teleman <stefan.teleman@gmail.com>
    
    This program is free software; you can redistribute it and/or
    modify it under the terms of version 2 of the GNU General Public
    License as published by the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
#include "config.h"

#include <kstat.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/sysinfo.h>
#include <sys/stat.h>
#include <vm/anon.h>
#include <errno.h>
#include <pthread.h>

#include "ksysguardd.h"
#include "Command.h"
#include "Common.h"
#include "Memory.h"

#ifdef PAGETOK
#undef PAGETOK
#endif
#define PAGETOK(a) (((t_memsize) sysconf(_SC_PAGESIZE)) * (t_memsize) a)

static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
static unsigned int exiting = 0;

static unsigned long interval = 0UL;
time_t ot = (time_t) 0;
time_t nt = (time_t) 0;
static vminfo_t ovminfo;
static vminfo_t nvminfo;
static uint32_t pagesize = 0UL;
static const unsigned long long mystery = 1048576ULL;
static uint32_t skip = 1;
static const t_memsize MB = (t_memsize)(1024ULL * 1024ULL);

static uint64_t physmem = 0ULL;
static t_memsize freemem = (t_memsize) 0;
static t_memsize totalmem = (t_memsize) 0;
static t_memsize reserved = (t_memsize) 0;
static t_memsize allocated = (t_memsize) 0;
static t_memsize available = (t_memsize) 0;
static t_memsize usedmem = (t_memsize) 0;
static t_memsize buffermem = (t_memsize) 0;
static t_memsize freeswap = (t_memsize) 0;
static t_memsize swapused = (t_memsize) 0;
static t_memsize totalswap = (t_memsize) 0;
static t_memsize pmem = (t_memsize) 0;
static uint64_t physpages = 0UL;
static uint64_t availrmem = 0ULL;
static uint64_t minfreemem = 0ULL;
static uint64_t pagesfreemem = 0ULL;
static uint64_t nfmem = 0ULL;
static uint32_t loop = 1;
static t_memsize ramcachemem = (t_memsize) 0;
static t_memsize cachedmem = (t_memsize) 0;
static t_memsize applmem = (t_memsize) 0;

static int Dirty = 1;
extern int errno;

static t_memsize
delta (t_memsize n, t_memsize o)
{
    return ((t_memsize) (n - o) * pagesize / mystery / interval);
}

static void
getVMInfo (t_memsize* freemem, t_memsize* reserved, t_memsize* allocated,
           t_memsize* available, t_memsize* freeswap)
{
    *freemem = delta (nvminfo.freemem, ovminfo.freemem);
    *reserved = delta (nvminfo.swap_resv, ovminfo.swap_resv);
    *allocated = delta (nvminfo.swap_alloc, ovminfo.swap_alloc);
    *available = delta (nvminfo.swap_avail, ovminfo.swap_avail);
    *freeswap = delta (nvminfo.swap_free, ovminfo.swap_free);
}

static void
setDefaultValues (void)
{
  physmem = (t_memsize) 0;
  freemem = (t_memsize) 0;
  totalmem = (t_memsize) 0;
  reserved = (t_memsize) 0;
  allocated = (t_memsize) 0;
  available = (t_memsize) 0;
  usedmem = (t_memsize) 0;
  buffermem = (t_memsize) 0;
  freeswap = (t_memsize) 0;
  swapused = (t_memsize) 0;
  totalswap = (t_memsize) 0;
  pmem = (t_memsize) 0;
  physpages = 0ULL;
  availrmem = 0ULL;
  minfreemem = 0ULL;
  pagesfreemem = 0ULL;
  nfmem = 0ULL;
  ramcachemem = (t_memsize) 0;
  cachedmem = (t_memsize) 0;
  applmem = (t_memsize) 0;
}

static void
setMemValues (void)
{
  totalmem = physmem;
  totalswap = freeswap + reserved;
  ramcachemem = totalmem;
  ramcachemem -= availrmem;
  swapused = allocated;
  buffermem = (reserved - allocated);
  cachedmem = ramcachemem + (reserved - allocated);
  usedmem = ramcachemem + swapused;
  physpages = nfmem / mystery;
  applmem = ramcachemem + allocated;
}

static void
processMemInfo (void)
{
  if (loop)
  {
    struct timeval tv;
    uint32_t i;

    (void) memset (&tv, '\0', sizeof (struct timeval));
    tv.tv_sec = 1;
    setDefaultValues();

    for (i = 0; i < 3; i++)
    {
      updateMemory();
      setMemValues();
      select (0L, NULL, NULL, NULL, &tv);
    }
    loop = 0;
  }
  else
  {
    updateMemory();
    setMemValues();
  }
  Dirty = 1;
}

static unsigned long
updateTime (void)
{
    ot = nt;
    nt = time(0);

    if ((time_t) 0 == ot)
        return 0UL;

    interval = (unsigned long) nt - ot;
    return interval;
}

t_memsize getPhysMem (void) { return physmem; }
t_memsize getFreeMem (void) { return freemem; }
t_memsize getTotalMem (void) { return totalmem; }
uint64_t getPhysPages (void) { return physpages; }
t_memsize getSwapReserved (void) { return reserved; }
t_memsize getSwapAllocated (void) { return allocated; }
t_memsize getSwapAvailable (void) { return available; }
t_memsize getSwapFree (void) { return freeswap; }
t_memsize getSwapTotal (void) { return totalswap; }
t_memsize getSwapUsed (void) { return swapused; }

void
initMemory (struct SensorModul* sm)
{
	registerMonitor("mem/physical/free", "integer", printMemFree, printMemFreeInfo, sm);
	registerMonitor("mem/physical/used", "integer", printMemUsed, printMemUsedInfo, sm);
  registerMonitor("mem/physical/application", "integer", printAppl, printApplInfo, sm);
  registerMonitor("mem/physical/buf", "integer", printBuffers, printBuffersInfo, sm);
  registerMonitor("mem/physical/cached", "integer", printCached, printCachedInfo, sm);
	registerMonitor("mem/swap/free", "integer", printSwapFree, printSwapFreeInfo, sm);
	registerMonitor("mem/swap/reserved", "integer", printSwapReserved, printSwapReservedInfo, sm);
	registerMonitor("mem/swap/used", "integer", printSwapUsed, printSwapUsedInfo, sm);
  updateMemory();
}

void exitMemory(void)
{
  pthread_mutex_lock (&mutex);
  exiting = 1;
  pthread_mutex_unlock (&mutex);

  removeMonitor("mem/physical/free");
  removeMonitor("mem/physical/used");
  removeMonitor("mem/physical/application");
  removeMonitor("mem/physical/buf");
  removeMonitor("mem/physical/cached");
  removeMonitor("mem/swap/free");
  removeMonitor("mem/swap/reserved");
  removeMonitor("mem/swap/used");
}

int updateMemory(void)
{
    kstat_ctl_t* kctl = 0L;
    kstat_t* ksp = 0L;
    kstat_t* kmp = 0L;
    kstat_named_t* knp = 0L;

    pthread_mutex_lock(&mutex);
    if (exiting == 1)
    {
      pthread_mutex_unlock(&mutex);
      return 0;
    }
    pagesize = sysconf (_SC_PAGESIZE);
    physmem = PAGETOK(sysconf(_SC_PHYS_PAGES)) / (1024ULL * 1024ULL);
    pmem = physmem;

    if (0UL == updateTime())
    {
      pthread_mutex_unlock(&mutex);
      return -1;
    }

    if ((kctl = kstat_open()) == (kstat_ctl_t *) 0L)
    {
        perror("kstat_open"), strerror(errno);
        pthread_mutex_unlock(&mutex);
        return -1;
    }

    while (kstat_chain_update(kctl) != 0);
    
    ksp = kstat_lookup (kctl, "unix", 0, "vminfo");
    if ((kstat_t *) 0L == ksp)
    {
        perror("kstat_lookup"), strerror(errno);
        kstat_close(kctl);
        pthread_mutex_unlock(&mutex);
        return -1;
    }
    if ((kid_t) -1 == kstat_read (kctl, ksp, &nvminfo))
    {
        perror("kstat_read"), strerror(errno);
        kstat_close(kctl);
        pthread_mutex_unlock(&mutex);
        return -1;
    }
    getVMInfo (&freemem, &reserved, &allocated, &available, &freeswap);
    ovminfo = nvminfo;
    if (skip)
    {
        skip = 0;
        kstat_close(kctl);
        pthread_mutex_unlock(&mutex);
        return 0;
    }

    while (kstat_chain_update(kctl) != 0);

    kmp = kstat_lookup (kctl, "unix", 0, "system_pages");
    if ((kstat_t *) 0L == kmp)
    {
      perror("kstat_lookup"), strerror(errno);
      kstat_close(kctl);
      pthread_mutex_unlock(&mutex);
      return -1;
    }

    if (kstat_read (kctl, kmp, 0L) == -1)
    {
      perror("kstat_read"), strerror(errno);
      kstat_close(kctl);
      pthread_mutex_unlock(&mutex);
      return -1;
    }

    while (kstat_chain_update(kctl) != 0);

    knp = (kstat_named_t *) kstat_data_lookup (kmp, "physmem");
    if (knp == (kstat_named_t *) 0L)
    {
      perror("kstat_data_lookup"), strerror(errno);
      kstat_close(kctl);
      pthread_mutex_unlock(&mutex);
      return -1;
    }
    physpages = knp->value.ui32 / mystery;
    // totalmem = PAGETOK(knp->value.ui32) / mystery;
    totalmem = physmem;
    nfmem = knp->value.ui32;

    while (kstat_chain_update(kctl) != 0);

    knp = (kstat_named_t *) kstat_data_lookup (kmp, "availrmem");
    if (knp == (kstat_named_t *) 0L)
    {
      perror("kstat_data_lookup"), strerror(errno);
      kstat_close(kctl);
      pthread_mutex_unlock(&mutex);
      return -1;
    }
    availrmem = PAGETOK(knp->value.ui32) / mystery;

    while (kstat_chain_update(kctl) != 0);

    knp = (kstat_named_t *) kstat_data_lookup (kmp, "minfree");
    if (knp == (kstat_named_t *) 0L)
    {
      perror("kstat_data_lookup"), strerror(errno);
      kstat_close(kctl);
      pthread_mutex_unlock(&mutex);
      return -1;
    }
    minfreemem = knp->value.ui32;

    while (kstat_chain_update(kctl) != 0);

    knp = (kstat_named_t *) kstat_data_lookup (kmp, "pagesfree");
    if (knp == (kstat_named_t *) 0L)
    {
      perror("kstat_data_lookup"), strerror(errno);
      kstat_close(kctl);
      pthread_mutex_unlock(&mutex);
      return -1;
    }
    pagesfreemem = knp->value.ui32;
    kstat_close(kctl);
	  Dirty = 1;
    pthread_mutex_unlock(&mutex);
	  return 1;
}

void printMemFree (const char* cmd)
{
    processMemInfo();
    (void) fprintf (CurrentClient, "%llu\n", freemem);
}

void printMemFreeInfo (const char* cmd)
{
    processMemInfo();
    (void) fprintf (CurrentClient, "Free Memory\t0\t%llu\tMB\n", totalmem);
}

void printMemUsed (const char* cmd)
{
    processMemInfo();
    (void) fprintf (CurrentClient, "%llu\n", usedmem);
}

void printMemUsedInfo (const char* cmd)
{
    processMemInfo();
    (void) fprintf (CurrentClient, "Used Memory\t0\t%llu\tMB\n", totalmem);
}

void printAppl (const char* cmd)
{
    processMemInfo();
    (void) fprintf (CurrentClient, "%llu\n", applmem);
}

void printApplInfo (const char* cmd)
{
    processMemInfo();
    (void) fprintf (CurrentClient, "Application Memory\t0\t%llu\tMB\n", totalmem);
}

void printBuffers (const char* cmd)
{
    processMemInfo();
    (void) fprintf (CurrentClient, "%llu\n", buffermem);
}

void printBuffersInfo (const char* cmd)
{
    processMemInfo();
    (void) fprintf (CurrentClient, "Buffer Memory\t0\t%llu\tMB\n", totalmem);
}

void printCached (const char* cmd)
{
    processMemInfo();
    (void) fprintf (CurrentClient, "%llu\n", cachedmem);
}

void printCachedInfo (const char* cmd)
{
    processMemInfo();
    (void) fprintf (CurrentClient, "Cached Memory\t0\t%llu\tMB\n", totalmem);
}

void printSwapUsed (const char* cmd)
{
    processMemInfo();
    (void) fprintf (CurrentClient, "%llu\n", swapused);
}

void printSwapUsedInfo (const char* cmd)
{
    processMemInfo();
    (void) fprintf (CurrentClient, "Used Swap Memory\t0\t%llu\tMB\n", totalswap);
}

void printSwapFree (const char* cmd)
{
    processMemInfo();
    (void) fprintf (CurrentClient, "%llu\n", freeswap);
}

void printSwapFreeInfo (const char* cmd)
{
    processMemInfo();
    (void) fprintf (CurrentClient, "Free Swap Memory\t0\t%llu\tMB\n", totalswap);
}

void printSwapReserved (const char* cmd)
{
    processMemInfo();
    (void) fprintf (CurrentClient, "%llu\n", reserved);
}

void printSwapReservedInfo (const char* cmd)
{
    processMemInfo();
    (void) fprintf (CurrentClient, "Reserved Swap Memory\t0\t%llu\tMB\n", totalswap);
}

// vim: ts=2 sw=2 et

