/*
  KSysGuard, the KDE System Guard
  
  Copyright (c) 2005 Stefan Teleman <stefan.teleman@gmail.com>
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of version 2 of the GNU General Public
  License as published by the Free Software Foundation.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  
*/

#include "SCSIStatCommon.h"

#ifdef __cplusplus
extern "C" {
#endif

extern int errno;

void
setUTSName (UTSName* utsName)
{
  if (uname (&(utsName->uName)) < 0)
  {
      perror("uname"), strerror(errno);
      utsName->ok = 0;
  }
  else
    utsName->ok = 1;
}

void
resetMnttabEntries (MnttabInfoCtrl** mntCtrlPtr)
{
  MnttabInfoCtrl* tmp;

  if (0L == mntCtrlPtr)
    return;

  tmp = *mntCtrlPtr;

  free (tmp->mnttabInfoPtr);
  tmp->mnttabInfoPtr = 0L;
  tmp->numberOfMnttabEntries = 0;
}

int
getMnttabInfo (MnttabInfoCtrl* ctrlPtr)
{
    struct extmnttab mp;
    FILE* fp = 0L;
    uint32_t i;

    if (0L == ctrlPtr)
        return -1;
    
    if ((fp = fopen (MNTTAB, "r")) == (FILE *) 0L)
    {
        perror("fopen"), strerror(errno);
        return 0L;
    }
    
    (void) memset ((void *) &mp, '\0', sizeof (struct extmnttab));
    resetMnttabEntries (&ctrlPtr);
    
    while (0 == getextmntent (fp, &mp, MNT_LINE_MAX))
    {
        if ((0 == strcmp (mp.mnt_fstype, "ufs")) || (0 == strcmp (mp.mnt_special, "swap")))
        {
            if (0L != addNewMnttabInfo (ctrlPtr, ctrlPtr->numberOfMnttabEntries))
            {
                (void) strcpy (ctrlPtr->mnttabInfoPtr[ctrlPtr->numberOfMnttabEntries].device, mp.mnt_special);
              (void) strcpy (ctrlPtr->mnttabInfoPtr[ctrlPtr->numberOfMnttabEntries].mountPoint, mp.mnt_mountp);
              (void) strcpy (ctrlPtr->mnttabInfoPtr[ctrlPtr->numberOfMnttabEntries].fsType, mp.mnt_fstype);
              ctrlPtr->mnttabInfoPtr[ctrlPtr->numberOfMnttabEntries].major = mp.mnt_major;
              ctrlPtr->mnttabInfoPtr[ctrlPtr->numberOfMnttabEntries].minor = mp.mnt_minor;
              ++(ctrlPtr->numberOfMnttabEntries);
            }
        }
    }
    resetmnttab (fp);
    fclose (fp);
    return (ctrlPtr->numberOfMnttabEntries ? 0 : -1);
}

void
fillDefaultValues (DiskSCSIInfo* dp)
{
  dp->instance = 0;
  dp->kid = 0;
  dp->lun = 0;
  dp->target = 0;
  dp->dev = 0x0;
  dp->major = 0;
  dp->minor = 0;
  (void) memset (dp->kstatName, '\0', sizeof (dp->kstatName));
  (void) memset (dp->kstatClass, '\0', sizeof (dp->kstatClass));
  (void) memset (dp->vendorName, '\0', sizeof (dp->vendorName));
  (void) memset (dp->productId, '\0', sizeof (dp->productId));
  (void) memset (dp->revision, '\0', sizeof (dp->revision));
  (void) memset (dp->ddiDeviceName, '\0', sizeof (dp->ddiDeviceName));
  (void) memset (dp->specType, '\0', sizeof (dp->specType));
  (void) memset (dp->nodeType, '\0', sizeof (dp->nodeType));
  (void) memset (dp->svr4Device, '\0', sizeof (dp->svr4Device));
  (void) memset (dp->scsiSlice, '\0', sizeof (dp->scsiSlice));
  (void) memset (dp->diMinorName, '\0', sizeof (dp->diMinorName));
  (void) memset (&(dp->kstatIO), '\0', sizeof (kstat_io_t));
}

void
copyDiskSCSIInfo (const kstat_t* ksp, const kstat_io_t* kip, DiskSCSIInfo* dp)
{
    if ((0L == ksp) || (0L == dp))
        return;
    
    dp->kid = ksp->ks_kid;
    (void) strcpy (dp->kstatName, ksp->ks_name);
    (void) strcpy (dp->kstatClass, ksp->ks_class);
    (void) memcpy (&(dp->kstatIO), kip, sizeof (kstat_io_t));
}

int
updateKStatList (DiskSCSIInfoCtrl* scsiCtrlPtr, InstanceCtrl* instanceCtrlPtr, const char* driver)
{
    DiskSCSIInfo* diskInfoPtr = 0L;
    kstat_ctl_t* kc = 0L;
    kstat_t* ks = 0L;
    kstat_io_t kio;
    uint32_t i;

    if ((0L == scsiCtrlPtr) || (0L == instanceCtrlPtr) || (0L == driver))
        return 0;

    if ((kc = kstat_open()) == (kstat_ctl_t *) NULL)
    {
        perror("kstat_open"), strerror(errno);
        return -1;
    }
    for (i = 0; i < instanceCtrlPtr->numInstances; ++i)
    {
      // ###
      (void) fprintf (stderr, "%s(): instance %ld\n", __func__, instanceCtrlPtr->instancePtr[i]);
      // ###
        if ((ks = kstat_lookup (kc, (char *) driver, instanceCtrlPtr->instancePtr[i], NULL)) == (kstat_t *) NULL)
        {
            perror("kstat_lookup"), strerror(errno);
            continue;
        }
        
        while (kstat_chain_update(kc) != 0);
        
        do
        {
            if ((strcmp (ks->ks_module, driver) == 0) && (ks->ks_type == KSTAT_TYPE_IO))
            {
                if (kstat_read (kc, ks, &kio) == -1)
                {
                    perror("kstat_read"), strerror(errno);
                    kstat_close(kc);
                    return -1;
                }
                diskInfoPtr = findSCSISlice (scsiCtrlPtr, ks->ks_name);
                if (0L == diskInfoPtr)
                {
                    while (kstat_chain_update(kc) != 0);
                    ks = ks->ks_next;
                    continue;
                }
                copyDiskSCSIInfo (ks, &kio, diskInfoPtr);
                while (kstat_chain_update(kc) != 0);
            }
            ks = ks->ks_next;
        } while (ks);
    }
    kstat_close(kc);
    return 0;
}

int
updateDeviceInfo (DiskSCSIInfoCtrl** scsiCtrlPtr, InstanceCtrl** instanceCtrlPtr, const char* driver)
{
    di_node_t rootNode;
    di_node_t node;
    di_prom_handle_t promHandle;
    DiskSCSICtrlNode* nodeCtrlPtr;
    int promProperties = 0;

    if ((0L == scsiCtrlPtr) || (0L == instanceCtrlPtr))
        return 0;

    if ((rootNode = di_init("/", DINFOCPYALL)) == DI_NODE_NIL)
    {
        perror("di_init"), strerror(errno);
        return -1;
    }
    if ((0 == getuid()) || (0 == geteuid()))
    {
        ++promProperties;
        if (promProperties && (promHandle = di_prom_init()) == DI_PROM_HANDLE_NIL)
        {
            perror("di_prom_init"), strerror(errno);
            di_fini(rootNode);
            return -1;
        }
    }
    nodeCtrlPtr = (DiskSCSICtrlNode *) malloc (sizeof (DiskSCSICtrlNode));
    if (0L == nodeCtrlPtr)
        return -1;
    
    node = di_drv_first_node (driver, rootNode);
    nodeCtrlPtr->scsiCtrlPtr = *scsiCtrlPtr;
    nodeCtrlPtr->instanceCtrlPtr = *instanceCtrlPtr;
    nodeCtrlPtr->node = node;
    nodeCtrlPtr->driver = driver;
    
    while (node != DI_NODE_NIL)
    {
        (void) di_walk_minor (node, NULL, 0, (void *) nodeCtrlPtr, dumpMinorDeviceNode);
        node = di_drv_next_node (node);
        nodeCtrlPtr->node = node;
    }
    
    if (promProperties)
        di_prom_fini(promHandle);
    
    di_fini(rootNode);
    free (nodeCtrlPtr);
    return 0;
}

void
resetInstanceInfo (InstanceCtrl* ctrlPtr, int** instancePtr)
{
    int* tmp;
    
    if (0L == ctrlPtr)
        return;
    
    if (ctrlPtr->instancePtr)
    {
        tmp = *instancePtr;
        free (tmp);
        *instancePtr = 0L;
    }
}


void
resetDDIInfo (DDIInfoCtrl* ctrlPtr, DDIInfo** ddiInfoPtr)
{
    DDIInfo* tmp;
    
    if (0L == ctrlPtr)
        return;
    
    if (ctrlPtr->ddiInfoPtr)
    {
        tmp = *ddiInfoPtr;
        free (tmp);
        *ddiInfoPtr = 0L;
    }
}

int
dumpMinorDeviceNode (di_node_t node, di_minor_t minor, void* arg)
{
    char* name;
    char* specType;
    char* nodeType;
    char* physPath;
    char* minorName;
    dev_t dev;
    DiskSCSIInfo* scsiInfo;
    DiskSCSICtrlNode* ctrlNode;
    int instance;
    char svr4[MAXPATHLEN];
    char tmpNode[MAXPATHLEN];
    char file[MAXPATHLEN];
    char partition[MAXPATHLEN];

    ctrlNode = (DiskSCSICtrlNode *) arg;
    if (0L == ctrlNode)
        return DI_WALK_TERMINATE;

    if (0L == ctrlNode->scsiCtrlPtr)
        return DI_WALK_TERMINATE;

    if (0L == ctrlNode->instanceCtrlPtr)
        return DI_WALK_TERMINATE;

    if ((caddr_t) node != (caddr_t) (ctrlNode->node))
        return DI_WALK_TERMINATE;

    name = di_devfs_path(node);
    
    instance = di_instance(node);
    // ###
    (void) fprintf (stderr, "%s(): instance = %ld\n", __func__, instance);
    // ###
    addInstance (ctrlNode->instanceCtrlPtr, &(ctrlNode->instanceCtrlPtr->instancePtr), instance);
    physPath = di_devfs_path(node);
    minorName = di_minor_name(minor);
    (void) memset (partition, '\0', MAXPATHLEN);
    (void) memset (svr4, '\0', MAXPATHLEN);
    
    (void) sprintf (tmpNode, "%s%d", ctrlNode->driver, instance);
    (void) strcpy (file, "/devices");
    (void) strcat (file, physPath);
    (void) strcat (file, ":");
    (void) strcat (file, minorName);
    
    specType = (di_minor_spectype(minor) == S_IFCHR) ? "char" : "block";
    nodeType = di_minor_nodetype(minor);
    
    if (di_minor_spectype(minor) == S_IFBLK)
        (void) strcpy (svr4, "/dev/dsk");
    else if (di_minor_spectype(minor) == S_IFCHR)
        (void) strcpy (svr4, "/dev/rdsk");
    else
    {
        di_devfs_path_free(name);
        return DI_WALK_TERMINATE;
    }
    if (0 == readDir(file, svr4, partition))
    {
        scsiInfo = findDiskSCSIInfo (ctrlNode->scsiCtrlPtr, file);
        if (0L == scsiInfo)
        {
            scsiInfo = addNewDiskSCSIInfo (&(ctrlNode->scsiCtrlPtr->scsiInfoPtr), ctrlNode->scsiCtrlPtr->numSCSIInfo);
            if (0L == scsiInfo)
            { 
                di_devfs_path_free(name);
                return DI_WALK_TERMINATE;
            }
            fillDefaultValues (&(ctrlNode->scsiCtrlPtr->scsiInfoPtr[ctrlNode->scsiCtrlPtr->numSCSIInfo]));
            scsiInfo = &(ctrlNode->scsiCtrlPtr->scsiInfoPtr[ctrlNode->scsiCtrlPtr->numSCSIInfo]);
            ++(ctrlNode->scsiCtrlPtr->numSCSIInfo);
        }
        scsiInfo->instance = instance;
        addInstance (ctrlNode->instanceCtrlPtr, &(ctrlNode->instanceCtrlPtr->instancePtr), scsiInfo->instance);
        scsiInfo->dev = di_minor_devt(minor);
        if (scsiInfo->dev != DDI_DEV_T_NONE)
        {
            scsiInfo->major = major(scsiInfo->dev);
            scsiInfo->minor = minor(scsiInfo->dev);
        }
        (void) strcpy (scsiInfo->ddiDeviceName, file);
        (void) strcpy (scsiInfo->svr4Device, partition);
        (void) strcpy (scsiInfo->specType, specType);
        (void) strcpy (scsiInfo->nodeType, ((nodeType) ? nodeType : ""));
        (void) strcpy (scsiInfo->diMinorName, di_minor_name(minor));
        (void) sprintf (scsiInfo->scsiSlice, "%s%d,%s", ctrlNode->driver, instance, scsiInfo->diMinorName);
        setProperties (node, scsiInfo);
        setTrickySCSIInfo (scsiInfo, ctrlNode->driver);
    } 
    di_devfs_path_free(name);
    return DI_WALK_CONTINUE;
}

void
setProperties (di_node_t node, DiskSCSIInfo* scsiInfo)
{
    int rval, i, type;
    di_prop_t prop, next;
    char buf[256];
    int* intp;
    char* strp;
    char* ptr;
    uchar_t* bytep;
    
    if ((next = di_prop_next(node, DI_PROP_NIL)) == DI_PROP_NIL)
        return;

    while (next != DI_PROP_NIL)
    {
        prop = next;
        next = di_prop_next(node, prop);
        type = di_prop_type(prop);
        if (type == DI_PROP_TYPE_UNKNOWN)
        {
            rval = di_prop_bytes(prop, &bytep);
            type = guessPropertyType(bytep, rval);
        }
        if (type == DI_PROP_TYPE_INT)
        {
            if ((strcmp (di_prop_name(prop), "lun") == 0) || (strcmp (di_prop_name(prop), "target") == 0))
            {
                rval = di_prop_ints(prop, &intp);
                if (rval == -1)
                {
                    perror("di_prop_ints"), strerror(errno);
                    break;
                }
                (void) memset (buf, '\0', sizeof(buf));
                for (i = 0; i < rval; ++i)
                {
                    ptr = &buf[i];
                    (void) sprintf (ptr, "%d", intp[i]);
                }
                if (strcmp(di_prop_name(prop), "lun") == 0)
                    scsiInfo->lun = atoi(buf);
                else
                    scsiInfo->target = atoi(buf); 
            }
        }
        else if (type == DI_PROP_TYPE_STRING)
        {
            if ((strcmp (di_prop_name(prop), "inquiry-vendor-id") == 0) ||
                (strcmp (di_prop_name(prop), "inquiry-product-id") == 0) ||
                (strcmp (di_prop_name(prop), "inquiry-revision-id") == 0))
            {
                rval = di_prop_strings(prop, &strp);
                if (rval == -1)
                {
                    perror("di_prop_strings"), strerror(errno);
                    break;
                }
                if (strcmp (di_prop_name(prop), "inquiry-vendor-id") == 0)
                    (void) sprintf (scsiInfo->vendorName, "%s", strp);
                else if (strcmp (di_prop_name(prop), "inquiry-product-id") == 0)
                    (void) sprintf (scsiInfo->productId, "%s", strp);
                else
                    (void) sprintf (scsiInfo->revision, "%s", strp);
                strp += strlen(strp) + 1;
            }
        }
    }
}

void
setTrickySCSIInfo (DiskSCSIInfo* scsiInfo, const char* driver)
{
    if (0L == scsiInfo)
        return;
    
    if (0 == strcmp (scsiInfo->specType, "block"))
    {
        if (0 == strcmp (scsiInfo->diMinorName, "h"))
        {
            (void) sprintf (scsiInfo->kstatName, "%s%d", driver, scsiInfo->target);
            if (0 == strlen(scsiInfo->kstatClass))
                (void) strcpy (scsiInfo->kstatClass, "disk");
        }
        else
        {
            if (0 == strlen(scsiInfo->kstatClass))
                (void) strcpy (scsiInfo->kstatClass, "disk");
        }
    }
}

void
setKIODiskStat (DiskSCSIInfo* scsiInfo)
{
  uint32_t i;
  int target;

  if (0L == scsiInfo)
    return;

  target = scsiInfo->target;

  scsiInfo->kstatIO.reads = 0;
  scsiInfo->kstatIO.writes = 0;
  scsiInfo->kstatIO.nread = 0ULL;
  scsiInfo->kstatIO.nwritten = 0ULL;
}


int
guessPropertyType(uchar_t *data, int len)
{
    int i, c, slen, guess;
    
    if (len < 0)
        return (-1);
    else if (len == 0)
        return (DI_PROP_TYPE_BOOLEAN);
    
    slen = 0;
    guess = DI_PROP_TYPE_STRING;
    
    for (i = 0; i < len; i++)
    {
        c = (int)data[i];
        switch (c)
        {
            case 0: /* NUL */
                if (slen == 0)
                    guess = DI_PROP_TYPE_BYTE;
                else
                    slen = 0;
                break;
            default:
                if (!isprint(c))
                    guess = DI_PROP_TYPE_BYTE;
                else
                    slen++;
        }
        if (guess != DI_PROP_TYPE_STRING)
            break;
    }
    if ((guess == DI_PROP_TYPE_BYTE) && (len % sizeof(int) == 0))
        guess = DI_PROP_TYPE_INT;
    
    return (guess);
}

int
readDir (const char* diskPath, const char* svr4Path, char* partitionName)
{
    DIR* dirp;
    struct dirent *dp;
    struct stat stbuf;
    char fileName[MAXPATHLEN];
    char tempName[MAXPATHLEN];
    char dirDname[MAXPATHLEN];
    
    if ((dirp = opendir(svr4Path)) != NULL)
    {
        while ((dp = readdir(dirp)) != NULL)
        {
            (void) sprintf(tempName, "../..%s", diskPath);
            (void) sprintf(dirDname, "%s/%s", svr4Path, dp->d_name);
            stat (dirDname, &stbuf);

            if (0 == strcmp (svr4Path, "/dev/dsk"))
            {
                if (S_ISBLK(stbuf.st_mode))
                {
                    readlink(dirDname, fileName, sizeof(fileName));
                    if (strcmp(fileName, tempName) == 0)
                    {
                        (void) strcpy (partitionName, dirDname);
                        closedir(dirp);
                        return 0;
                    }
                }
            }
            else if (0 == strcmp (svr4Path, "/dev/rdsk"))
            {
                if (S_ISCHR(stbuf.st_mode))
                {
                    readlink(dirDname, fileName, sizeof(fileName));
                    if (strcmp(fileName, tempName) == 0)
                    {
                        (void) strcpy (partitionName, dirDname);
                        closedir(dirp);
                        return 0;
                    }
                }
            }
            (void) memset (fileName,'\0', MAXPATHLEN);
            (void) memset (tempName,'\0', MAXPATHLEN);
            (void) memset (dirDname,'\0', MAXPATHLEN);
        }
        closedir(dirp);
    }
    return(-1);
}

DiskSCSIInfo*
addNewDiskSCSIInfo (DiskSCSIInfo** scsiInfoPtr, uint32_t currentSize)
{
  DiskSCSIInfo* ptr;
  DiskSCSIInfo* tmp;
  uint32_t i;

  if (0L == scsiInfoPtr)
      return 0L;
  
  ptr = (DiskSCSIInfo *) malloc ((1 + currentSize) * sizeof (DiskSCSIInfo));
  if (0L == ptr)
  {
    perror("malloc"), strerror(errno);
    return 0L;
  }

  if (0 == currentSize)
  {
      *scsiInfoPtr = ptr;
  }
  else
  {
      tmp = (*scsiInfoPtr);
      for (i = 0; i < (uint32_t) currentSize; ++i)
          (void) memcpy (&(ptr[i]), &(tmp[i]), sizeof (DiskSCSIInfo));
      
      tmp = (*scsiInfoPtr);
      (*scsiInfoPtr) = ptr;
      free (tmp);
  }
  return &((*scsiInfoPtr)[currentSize]);
}

MnttabInfo*
addNewMnttabInfo (MnttabInfoCtrl* ctrlPtr, uint32_t currentSize)
{
    MnttabInfo* ptr = 0L;
    MnttabInfo* tmp = 0L;
    uint32_t i;

    if (0L == ctrlPtr)
        return 0L;
    
  ptr = (MnttabInfo *) malloc ((1 + currentSize) * sizeof (MnttabInfo));
  if (0L == ptr)
  {
    perror("malloc"), strerror(errno);
    return 0L;
  }

  if (0 == currentSize)
  {
      ctrlPtr->mnttabInfoPtr = ptr;
  }
  else
  {
      for (i = 0; i < currentSize; ++i)
      {
        (void) memset (&(ptr[i]), '\0', sizeof (MnttabInfo));
        (void) memcpy (&(ptr[i]), &(ctrlPtr->mnttabInfoPtr[i]), sizeof (MnttabInfo));
      }
      tmp = ctrlPtr->mnttabInfoPtr;
      ctrlPtr->mnttabInfoPtr = ptr;
      free (tmp);
  }
  return (&(ctrlPtr->mnttabInfoPtr[currentSize]));
}


DiskSCSIInfo*
findDiskSCSIInfo (DiskSCSIInfoCtrl* ctrlPtr, const char* ddiDeviceName)
{
    uint32_t i;
    
    if ((0L == ddiDeviceName) || (0L == ctrlPtr))
        return 0L;
    
    for (i = 0; i < ctrlPtr->numSCSIInfo; ++i)
    {
        if (0 == strcmp (ctrlPtr->scsiInfoPtr[i].ddiDeviceName, ddiDeviceName))
            return (&(ctrlPtr->scsiInfoPtr[i]));
    }
    return 0L;
}

DDIInfo*
findDDIInfo (DDIInfoCtrl* ctrlPtr, const char* ddiDeviceName)
{
    uint32_t i;

    if ((0L == ddiDeviceName) || (0L == ctrlPtr))
        return 0L;
    
    for (i = 0; i < ctrlPtr->numDDIInfo; ++i)
    {
        if (0 == strcmp (ctrlPtr->ddiInfoPtr[i].physicalPath, ddiDeviceName))
            return (&(ctrlPtr->ddiInfoPtr[i]));
    }
    return 0L;
}

int
findInstance (InstanceCtrl* ctrlPtr, int instance)
{
    uint32_t i;

    if (0L == ctrlPtr)
        return -1;
    
    for (i = 0; i < ctrlPtr->numInstances; ++i)
        if (ctrlPtr->instancePtr[i] == instance)
            return ctrlPtr->instancePtr[i];

    return -1;
}

DiskSCSIInfo*
findSCSISlice (DiskSCSIInfoCtrl* ctrlPtr, const char* sliceName)
{
    uint32_t i;
    
    if ((0L == sliceName) || (0L == ctrlPtr))
        return 0L;
    
    for (i = 0; i < ctrlPtr->numSCSIInfo; ++i)
        if (strcmp (ctrlPtr->scsiInfoPtr[i].scsiSlice, sliceName) == 0)
            return (&(ctrlPtr->scsiInfoPtr[i]));

    return 0L;
}

MnttabInfo*
findMnttabInfo (MnttabInfoCtrl* ctrlPtr, int32_t major, int32_t minor)
{
    uint32_t i;

    if (0L == ctrlPtr)
        return 0L;
    
    for (i = 0; i < ctrlPtr->numberOfMnttabEntries; ++i)
    {
        if ((ctrlPtr->mnttabInfoPtr[i].major == major) &&
            (ctrlPtr->mnttabInfoPtr[i].minor == minor))
            return (&(ctrlPtr->mnttabInfoPtr[i]));
    }
    return 0L;
}


void
addInstance (InstanceCtrl* ctrlPtr, int** instancePtr, int instance)
{
  if ((0L == instancePtr) || (0L == ctrlPtr))
      return;
    
  if (findInstance (ctrlPtr, instance) < 0)
  {
    (*instancePtr) = (int *) realloc ((*instancePtr), (sizeof(int) * (1 + ctrlPtr->numInstances)));
    if (*instancePtr)
    {
       (*instancePtr)[ctrlPtr->numInstances] = instance;
       ++(ctrlPtr->numInstances);
    }
  }
}

#ifdef __cplusplus
} // extern "C"
#endif

// vim: ts=2 sw=2 et
