/*
    KSysGuard, the KDE System Guard
    
	  Copyright (c) 2005 Stefan Teleman <stefan.teleman@gmail.com>
    
    This program is free software; you can redistribute it and/or
    modify it under the terms of version 2 of the GNU General Public
    License as published by the Free Software Foundation.
    
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#ifndef KSG_SCSISTATCOMMON_H
#define KSG_SCSISTATCOMMON_H

#include <stdio.h>
#include <sys/types.h>
#include <sys/dkio.h>
#include <sys/stat.h>
#include <sys/mntent.h>
#include <sys/mnttab.h>
#include <sys/statvfs.h>
#include <sys/utsname.h>
#include <string.h>
#include <strings.h>
#include <unistd.h>
#include <stdlib.h>
#include <ctype.h>
#include <inttypes.h>
#include <kstat.h>
#include <fcntl.h>
#include <sys/fcntl.h>
#include <sys/param.h>
#include <sys/mkdev.h>
#include <sys/sysinfo.h>
#include <devid.h>
#include <libdevinfo.h>
#include <dirent.h>
#include <errno.h>

#ifdef __cplusplus
extern "C" {
#endif

  // DDI/DKI stuff
typedef struct DDIInfo
{
  int instance;
  char driver[8];
  char physicalPath[MAXPATHLEN];
} DDIInfo;

typedef struct DDIInfoCtrl
{
  DDIInfo* ddiInfoPtr;
  uint32_t numDDIInfo;
} DDIInfoCtrl;
    
// KStat stuff
typedef struct DiskIOInfo
{
  kstat_io_t old;
  kstat_io_t new;
} DiskIOInfo;

// SCSI stuff
typedef struct DiskSCSIInfo
{
    int instance;
    int kid;
    int lun;
    int target;
    dev_t dev;
    major_t major;
    minor_t minor;
    char kstatName[KSTAT_STRLEN];
    char kstatClass[KSTAT_STRLEN];
    char vendorName[32];
    char productId[64];
    char revision[32];
    char ddiDeviceName[MAXPATHLEN];
    char specType[16];
    char nodeType[64];
    char svr4Device[MAXPATHLEN];
    char scsiSlice[32];
    char diMinorName[2];
    kstat_io_t kstatIO;
} DiskSCSIInfo;

typedef struct DiskSCSIInfoCtrl
{
    DiskSCSIInfo* scsiInfoPtr;
    uint32_t numSCSIInfo;
} DiskSCSIInfoCtrl;

typedef struct MnttabInfo
{
    char device[MAXPATHLEN + sizeof(char)];
    char mountPoint[MAXPATHLEN + sizeof(char)];
    char fsType[FSTYPSZ + sizeof(char)];
    int32_t major;
    int32_t minor;
} MnttabInfo;

typedef struct MnttabInfoCtrl
{
    MnttabInfo* mnttabInfoPtr;
    uint32_t numberOfMnttabEntries;
} MnttabInfoCtrl;

typedef struct InstanceCtrl
{
    int* instancePtr;
    uint32_t numInstances;
} InstanceCtrl;

typedef struct DiskSCSICtrlNode
{
    const char* driver;
    DiskSCSIInfoCtrl* scsiCtrlPtr;
    InstanceCtrl* instanceCtrlPtr;
    di_node_t node;
} DiskSCSICtrlNode;

// NFS stuff
typedef struct NFSMountInfo
{
    char mnt_special[MAXPATHLEN];
    char mnt_mountp[MAXPATHLEN];
    char mnt_mntopts[MNT_LINE_MAX];
    uint_t mnt_major;
    uint_t mnt_minor;
} NFSMountInfo;

typedef struct NFSMountInfoCtrl
{
    NFSMountInfo* nfsInfoPtr;
    uint32_t numberOfNFSMountEntries;
} NFSMountInfoCtrl;

// UTSName stuff
typedef struct UTSName
{
  int ok;
  struct utsname uName;
} UTSName;  

void setUTSName (UTSName* utsName);
/* void setDriverName (const char* driver); */
void fillDefaultValues (DiskSCSIInfo* dsp);
void copyDiskSCSIInfo (const kstat_t* ksp, const kstat_io_t* kip, DiskSCSIInfo* dsp);
int getMnttabInfo (MnttabInfoCtrl* mntCtrlPtr);
void resetMnttabEntries (MnttabInfoCtrl** mntCtrlPtr);
int dumpDeviceNode (di_node_t node, void* arg);
int dumpMinorDeviceNode (di_node_t node, di_minor_t minor, void *arg);
void walkDiskNodes (const char* driver, di_node_t node);
int guessPropertyType(uchar_t *data, int len);
int updateDeviceAlias (const char* driver);
int updateKStatList (DiskSCSIInfoCtrl* scsiCtrlPtr, InstanceCtrl* instanceCtrlPtr, const char* driver);
int updateDeviceInfo (DiskSCSIInfoCtrl** scsiCtrlPtr, InstanceCtrl** instanceCtrlPtr, const char* driver);
void updateDDIInfo (di_node_t node, const char* driver);
int readDir (const char* diskPath, const char* svr4Path, char* partitionName);
void setProperties (di_node_t node, DiskSCSIInfo* scsiInfo);
void setTrickySCSIInfo (DiskSCSIInfo* scsiInfo, const char* driver);
void setKIODiskStat (DiskSCSIInfo* scsiInfo);
DiskSCSIInfo* findDiskSCSIInfo (DiskSCSIInfoCtrl* ctrlPtr, const char* ddiDeviceName);
DiskSCSIInfo* addNewDiskSCSIInfo (DiskSCSIInfo** scsiInfoPtr, uint32_t currentSize);
MnttabInfo* addNewMnttabInfo (MnttabInfoCtrl* ctrlPtr, uint32_t currentSize);
DDIInfo* findDDIInfo (DDIInfoCtrl* ctrlPtr, const char* ddiDeviceName);
int findInstance (InstanceCtrl* ctrlPtr, int instance);
DiskSCSIInfo* findSCSISlice (DiskSCSIInfoCtrl* ctrlPtr, const char* sliceName);
MnttabInfo* findMnttabinfo (MnttabInfoCtrl* ctrlPtr, int32_t major, int32_t minor);
void addInstance (InstanceCtrl* ctrlPtr, int** instancePtr, int instance);
void resetInstanceInfo (InstanceCtrl* ctrlPtr, int** instancePtr);

#ifdef __cplusplus
} // extern "C"
#endif

#endif // KSG_SCSISTATCOMMON_H

// vim: ts=2 sw=2 et
